/*
 *	wiiuse
 *
 *	Written By:
 *		Michael Laforest	< para >
 *		Email: < thepara (--AT--) g m a i l [--DOT--] com >
 *
 *	Copyright 2006-2007
 *
 *	This file is part of wiiuse.
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 3 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *	$Header$
 *
 */

/**
 *	@file
 *	@brief Handles device connections for Windows.
 *
 *	The file includes functions that handle connections
 *	to the wiimote device(s).
 */

#ifdef WIN32

#include <stdio.h>
#include <malloc.h>

#include <windows.h>
#include <hidsdi.h>
#include <setupapi.h>

#include "definitions.h"
#include "wiiuse_internal.h"
#include "connect.h"

static int wiiuse_connect_single(struct wiimote_t* wm, char* address);

int wiiuse_find(struct wiimote_t** wm, int max_wiimotes, int timeout) {
	GUID device_id;
	HANDLE dev;
	HDEVINFO device_info;
	int i, index, len;
	SP_DEVICE_INTERFACE_DATA device_data;
	PSP_DEVICE_INTERFACE_DETAIL_DATA detail_data;
	HIDD_ATTRIBUTES	attr;
	int found_wiimotes = 0;

	device_data.cbSize = sizeof(device_data);
	index = 0;

	/* get the device id */
	HidD_GetHidGuid(&device_id);

	/* query device info */
	device_info = SetupDiGetClassDevs(&device_id, NULL, NULL, (DIGCF_INTERFACEDEVICE | DIGCF_PRESENT));

	do {
		/* query the next device info */
		i = SetupDiEnumDeviceInterfaces(device_info, NULL, &device_id, index, &device_data);

		if (!i)
			break;

		/* get the size of the data block required */
		i = SetupDiGetDeviceInterfaceDetail(device_info, &device_data, NULL, 0, &len, NULL);
		detail_data = malloc(len);
		detail_data->cbSize = sizeof(SP_DEVICE_INTERFACE_DETAIL_DATA);

		/* query the data for this device */
		i = SetupDiGetDeviceInterfaceDetail(device_info, &device_data, detail_data, len, NULL, NULL);

		/* open the device */
		dev = CreateFile(detail_data->DevicePath, 0, (FILE_SHARE_READ | FILE_SHARE_WRITE), (LPSECURITY_ATTRIBUTES)NULL, OPEN_EXISTING,	0, NULL);

		/* get device attributes */
		attr.Size = sizeof(attr);
		i = HidD_GetAttributes(dev, &attr);

		if ((attr.VendorID == WM_VENDOR_ID) && (attr.ProductID == WM_PRODUCT_ID)) {
			/* this is a wiimote */
			wm[found_wiimotes]->device_path = _strdup(detail_data->DevicePath);
			wm[found_wiimotes]->device_handle = dev;

			INFO("Found wiimote [assigned wiimote id %i].", wm[found_wiimotes]->unid);

			WIIMOTE_ENABLE_STATE(wm[found_wiimotes], WIIMOTE_STATE_DEV_FOUND);

			++found_wiimotes;
			if (found_wiimotes >= max_wiimotes)
				return found_wiimotes;
		} else {
			/* not a wiimote */
			CloseHandle(dev);
		}

		free(detail_data);

		++index;

	} while(1);

	SetupDiDestroyDeviceInfoList(device_info);

	return found_wiimotes;
}


int wiiuse_connect(struct wiimote_t** wm, int wiimotes) {
	int connected = 0;
	int i = 0;

	for (; i < wiimotes; ++i) {
		if (!WIIMOTE_IS_SET(wm[i], WIIMOTE_STATE_DEV_FOUND))
			/* if the device address is not set, skip it */
			continue;

		if (wiiuse_connect_single(wm[i], NULL))
			++connected;
	}

	return connected;
}


static int wiiuse_connect_single(struct wiimote_t* wm, char* address) {
	if (!wm || WIIMOTE_IS_CONNECTED(wm))
		return 0;

	wm->sock = CreateFile(wm->device_path,
						(GENERIC_READ | GENERIC_WRITE),
						(FILE_SHARE_READ | FILE_SHARE_WRITE),
						NULL, OPEN_EXISTING, FILE_FLAG_OVERLAPPED, NULL);
	if (wm->sock == INVALID_HANDLE_VALUE)
		return 0;

	wm->hid_overlap.hEvent = CreateEvent(NULL, 1, 1, "");
	wm->hid_overlap.Offset = 0;
	wm->hid_overlap.OffsetHigh = 0;

	/* connected, but no handshake yet */
	WIIMOTE_ENABLE_STATE(wm, WIIMOTE_STATE_CONNECTED);
	WIIMOTE_ENABLE_STATE(wm, WIIMOTE_STATE_HANDSHAKE);
	wiiuse_handshake(wm, NULL, 0);

	wiiuse_set_leds(wm, WIIMOTE_LED_NONE);

	return 1;
}


void wiiuse_disconnect(struct wiimote_t* wm) {
	if (!wm || WIIMOTE_IS_CONNECTED(wm))
		return;

	CloseHandle(wm->sock);
	CloseHandle(wm->device_handle);
	wm->sock = 0;

	free(wm->device_path);

	ResetEvent(&wm->hid_overlap);

	WIIMOTE_DISABLE_STATE(wm, WIIMOTE_STATE_CONNECTED);
	WIIMOTE_DISABLE_STATE(wm, WIIMOTE_STATE_HANDSHAKE);
}


void wiiuse_handshake(struct wiimote_t* wm, byte* data, unsigned short len) {
	if (!wm)	return;

	switch (wm->handshake_state) {
		case 0:
		{
			/* send request to wiimote for accelerometer calibration */
			byte* buf = (byte*)malloc(sizeof(byte) * 8);
			wiiuse_read_data(wm, wiiuse_handshake, buf, WM_MEM_OFFSET_CALIBRATION, 7);
			wm->handshake_state++;

			wiiuse_set_leds(wm, WIIMOTE_LED_NONE);
			break;
		}
		case 1:
		{
			struct read_req_t* req = wm->read_req;
			struct accel_t* accel = &wm->accel_calib;

			/* received read data */
			accel->cal_zero.x = req->buf[0];
			accel->cal_zero.y = req->buf[1];
			accel->cal_zero.z = req->buf[2];

			accel->cal_g.x = req->buf[4];
			accel->cal_g.y = req->buf[5];
			accel->cal_g.z = req->buf[6];

			/* done with the buffer */
			free(req->buf);

			/* handshake is done */
			DEBUG("Handshake finished. Calibration: Idle: X=%x Y=%x Z=%x\t+1g: X=%x Y=%x Z=%x",
					accel->cal_zero.x, accel->cal_zero.y, accel->cal_zero.z,
					accel->cal_g.x, accel->cal_g.y, accel->cal_g.z);

			WIIMOTE_DISABLE_STATE(wm, WIIMOTE_STATE_HANDSHAKE);
			wm->handshake_state++;
			break;
		}
		default:
		{
			break;
		}
	}
}

#endif
